/*
' COPanno Optimized rectangular cutting tool
'
'
' Copyright (C)2012-2015 Florian Foinant
'
' This program is free software; you can redistribute it and/or modify
' it under the terms of the GNU General Public License as published by
' the Free Software Foundation; either version 2 of the License, or
' (at your option) any later version.
'
' This program is distributed in the hope that it will be useful,
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
' GNU General Public License for more details.
'
' You should have received a copy of the GNU General Public License
' along with this program; if not, write to the Free Software
' Foundation, Inc., 51 Franklin St, Fifth Floor,
' Boston, MA  02110-1301  USA
'*/

import java.util.ArrayList;

/* Algo de calcul du découpage : brute-force recursif */
class Algo {

	/**
	 *
	 * @param Feuille : Dimensions des feuilles disponibles
	 * @param Debit : Liste des panneaux à débiter
	 * @param quick : Utiliser la version simplifier de l'algorithme, rapide mais la solution n'est pas forcement la meilleure
	 * @return L'ensemble des panneaux positionnées sur des feuilles de la dimension spécifiée.
	 */
	public static Solution Resoudre(Panneau Feuille, ArrayList<Panneau> Debit, boolean quick){
		return Algo.Resoudre(Feuille, Debit, quick, 0);
	}
	
	/**
	 *
	 * @param cas : A quel cas restreindre le calcul ?
	 **/ 
	public static Solution Resoudre(Panneau Feuille, ArrayList<Panneau> Debit, boolean quick, int cas){

		Solution solution = new Solution();	//résultat de la fonction
		Solution solutionp = new Solution(); //variable pour la boucle
		PanneauPositionne feuille_entiere;
		
		//Initialisation de la liste de proposition
		solution.proposition = new ArrayList<PanneauPositionne>();
		//initialisation de la liste des chutes
		solution.chutes = new ArrayList<PanneauPositionne>();


		//Initialisation de la feuille à couper
		feuille_entiere = new PanneauPositionne(Feuille.largeur, Feuille.hauteur);
		feuille_entiere.position = new Point(0, 0);
		
		//Il reste tous les panneaux à couper, on les trie par ordre de surface
		solutionp.reste = Panneau.TriSurface(Debit);
		
		//Tant qu'il reste des panneaux à couper et (garde-fou) qu'il y a moins de 100 feuilles utilisées
		while (solutionp.reste.size() > 0 && feuille_entiere.numFeuille<100){
			//On positionne tous les panneaux possible dans une feuille
			solutionp.proposition = new ArrayList<PanneauPositionne>();
			solutionp = RecResoudre(feuille_entiere, solutionp, quick, cas);
			
			solution.chutes.addAll(solutionp.chutes);
			if (solutionp.proposition.size()>0)
				//on transfert les positionnements dans le resultat
				solution.proposition.addAll(solutionp.proposition);
			else{
				//aucun des panneaux restant ne passe dans la feuille
				for(int n=0 ; n<solutionp.reste.size(); n++){
					//on transfert le reste dans le resultat final en le déclarant hors gabarit 
					solution.proposition.add(solutionp.reste.get(n).horsFeuille());
				}
				break;
			}
			//on passe à la feuille suivante
			feuille_entiere.numFeuille ++;
		}
		
		return solution;
	}

	/**Ajoute un panneau à une solution existante */
	private static Solution PreparerSolution (Solution solution_1, PanneauPositionne sol){

		Solution nouvelleSolution = new Solution();
		
		//on transfert tous les panneaux déjà positionnés
		nouvelleSolution.proposition.addAll(solution_1.proposition);
		nouvelleSolution.proposition.add(sol);
		
		//pour chaque panneau restant
		for (int n=0 ; n<solution_1.reste.size() ; n++){
			Panneau tmp = solution_1.reste.get(n).clone();
			//si il passe
			if ((tmp.largeur==sol.largeur && tmp.hauteur==sol.hauteur)
			|| (tmp.largeur==sol.hauteur && tmp.hauteur==sol.largeur && ! tmp.fil)){
				//on en déduit un du reste
				if (tmp.nombre>1){
					tmp.nombre--;
					nouvelleSolution.reste.add(tmp);
				}else{
					solution_1.reste.remove(n);
				}
			}else{
				//s'il ne passe pas, il reste dans la nouvelle solution
				nouvelleSolution.reste.add(tmp);
			}
		}

		return nouvelleSolution;

	}
	
	/**Raccourci pour RecResoudre avec calcul de tous les cas*/
	private static Solution RecResoudre(PanneauPositionne Feuille, Solution solution, boolean quick){
		return RecResoudre(Feuille,solution,quick,0);
	}
			
	/**
	 * Calcul récursif
	 * @param Feuille : la feuille ou la chute dans laquelle positionner les panneaux
	 * @param solution : contenant pour les panneaux
	 * @param quick : utilisation de la méthode simplifiée
	 * @param cas : définie par laquelle des huit branches il faut commencer la recherche, 0 pour toutes
	 */
	private static Solution RecResoudre(PanneauPositionne Feuille, Solution solution, boolean quick, int cas){

		PanneauPositionne chute1;
		PanneauPositionne chute2;
		PanneauPositionne sol;

		//Les 8 branches de l'arbre des possibilités
		ArrayList<Solution> nsolution = new ArrayList<Solution>(8);
		Solution currSol;

		Panneau pan;
		boolean panpasse;

		//Feuille de surface nulle : c'est une chute qui a été completement remplie
		//on a plus rien à faire avec
		if(Feuille.getSurface() == 0){
			return solution;
		}

		//Tous les panneaux sont placés, la feuille actuelle est une chute
		if(solution.reste.size() == 0){
			solution.chutes.add(Feuille);
			return solution;
		}
		
		//Le plus petit panneau ne passe pas, la feuille actuelle est une chute inutilisable
		pan = solution.reste.get(solution.reste.size()-1);
		if(pan.getSurface() > Feuille.getSurface()){
			solution.chutes.add(Feuille);
			return solution;
		}

		//On teste les panneaux du plus grand au plus petit
		for(int n=0 ; n<solution.reste.size() ; n++){
			pan = solution.reste.get(n);
			panpasse = false;

			//panneau trop grand pour la feuille, on passe au suivant
			if (pan.getSurface() > Feuille.getSurface())
				continue;
			
			//la feuille est considérée horizontale = largeur en x, hauteur en y

			//si le panneau entre horizontalement
			if (pan.largeur <= Feuille.largeur && pan.hauteur <= Feuille.hauteur) {
				sol = new PanneauPositionne(pan.largeur, pan.hauteur, pan.repere, Feuille.numFeuille);
				sol.position.set(Feuille.position);
				
				//Premier cas de chute : coupe verticale à travers
				chute1 = new PanneauPositionne(pan.largeur, Feuille.hauteur - pan.hauteur, "", Feuille.numFeuille);
				chute1.position.x = Feuille.position.x;
				chute1.position.y = Feuille.position.y + pan.hauteur;
				chute2 = new PanneauPositionne(Feuille.largeur - pan.largeur, Feuille.hauteur, "", Feuille.numFeuille);
				chute2.position.x = Feuille.position.x + pan.largeur;
				chute2.position.y = Feuille.position.y;
				
				if (cas==0 || cas==1){
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute1, currSol, quick);
					currSol=RecResoudre(chute2, currSol, quick);
					nsolution.add(currSol);
				}
				if (! quick && (cas==0 || cas==2)){
					//la même chose en inversant l'ordre dans lequel on teste les chutes
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute2, currSol, quick);
					currSol=RecResoudre(chute1, currSol, quick);
					nsolution.add(currSol);
				}
				
				//Second cas de chute : coupe horizontale à travers
				chute1 = new PanneauPositionne(Feuille.largeur, Feuille.hauteur - pan.hauteur, "", Feuille.numFeuille);
				chute1.position.x = Feuille.position.x;
				chute1.position.y = Feuille.position.y + pan.hauteur;
				chute2 = new PanneauPositionne(Feuille.largeur - pan.largeur, pan.hauteur, "", Feuille.numFeuille);
				chute2.position.x = Feuille.position.x + pan.largeur;
				chute2.position.y = Feuille.position.y;
				
				if (cas==0 || cas==3){
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute1, currSol, quick);
					currSol=RecResoudre(chute2, currSol, quick);
					nsolution.add(currSol);
				}

				if (! quick && (cas==0 || cas==4)){
					//la même chose en inversant l'ordre dans lequel on teste les chutes
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute2, currSol, quick);
					currSol=RecResoudre(chute1, currSol, quick);
					nsolution.add(currSol);
				}

				panpasse = true;
			}

			//si le panneau entre verticalement et qu'on ne garde pas le fil
			if(pan.largeur <= Feuille.hauteur && pan.hauteur <= Feuille.largeur && ! pan.fil) {
				sol = new PanneauPositionne(pan.hauteur, pan.largeur, pan.repere, Feuille.numFeuille);
				sol.position.set(Feuille.position);

				//Premier cas de chute : coupe verticale à travers
				chute1 = new PanneauPositionne(pan.hauteur, Feuille.hauteur - pan.largeur, "", Feuille.numFeuille);
				chute1.position.x = Feuille.position.x;
				chute1.position.y = Feuille.position.y + pan.largeur;
				chute2 = new PanneauPositionne(Feuille.largeur - pan.hauteur, Feuille.hauteur, "", Feuille.numFeuille);
				chute2.position.x = Feuille.position.x + pan.hauteur;
				chute2.position.y = Feuille.position.y;
				
				if (cas==0 || cas==5){
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute1, currSol, quick);
					currSol=RecResoudre(chute2, currSol, quick);
					nsolution.add(currSol);
				}

				if (! quick && (cas==0 || cas==6)){
					//la même chose en inversant l'ordre dans lequel on teste les chutes
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute2, currSol, quick);
					currSol=RecResoudre(chute1, currSol, quick);
					nsolution.add(currSol);
				}

				//Second cas de chute : coupe horizontale à travers
				chute1 = new PanneauPositionne(Feuille.largeur, Feuille.hauteur - pan.largeur, "", Feuille.numFeuille);
				chute1.position.x = Feuille.position.x;
				chute1.position.y = Feuille.position.y + pan.largeur;
				chute2 = new PanneauPositionne(Feuille.largeur - pan.hauteur, pan.largeur, "", Feuille.numFeuille);
				chute2.position.x = Feuille.position.x + pan.hauteur;
				chute2.position.y = Feuille.position.y;

				if (cas==0 || cas==7){
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute1, currSol, quick);
					currSol=RecResoudre(chute2, currSol, quick);
					nsolution.add(currSol);
				}

				if (! quick && (cas==0 || cas==8)){
					//la même chose en inversant l'ordre dans lequel on teste les chutes
					currSol=PreparerSolution(solution, sol);
					currSol=RecResoudre(chute2, currSol, quick);
					currSol=RecResoudre(chute1, currSol, quick);
					nsolution.add(currSol);
				}

				panpasse = true;
			}

			if(panpasse)
				break;
		}

		if (nsolution.size() == 0){
			//aucun panneau ne passe
			solution.chutes.add(Feuille);
			return solution;
		}

		int maxi2 = FiltreSolution(nsolution);

		if(solution.chutes != null)
			nsolution.get(maxi2).chutes.addAll(solution.chutes);
		return nsolution.get(maxi2);
	}
	
	/**
	 * Trouve la meilleure solution dans la liste fournie
	 * @param nSolution Liste des solutions à filtrer
	 * @return l'indice de la meilleure solution
	 */
	public static int FiltreSolution(ArrayList<Solution> nSolution){
		//choix de la solution qui fait entrer le plus de panneaux
		int maxi = 0;
		for (int n = 0 ; n<nSolution.size() ; n++){
			if ( nSolution.get(n) != null){
				if (nSolution.get(maxi).proposition.size() < nSolution.get(n).proposition.size()) {
					maxi = n;
				}else if (nSolution.get(maxi) == null) {
					maxi = n + 1;
				}
			}
		}

		//Choix de la solution qui fait le moins de chute parmis celles qui font entrer le plus de panneaux
		int mini = maxi;
		for (int n=0 ; n<nSolution.size() ; n++){
			if (nSolution.get(n) != null) {
				if (nSolution.get(n).proposition.size() == nSolution.get(maxi).proposition.size()) {
					if (nSolution.get(mini).chutes.size() > nSolution.get(n).chutes.size())
						mini = n;
				}
			}
		}

		//Choix de la solution qui fait la chute la plus grande parmis celles qui font le moins de chute
		int maxi2 = mini;
		double maxisurf = 0;
		for (int n=0 ; n<nSolution.size() ; n++)
			if (nSolution.get(n) != null)
				if (nSolution.get(n).chutes.size() == nSolution.get(mini).chutes.size()
				&& nSolution.get(n).proposition.size() == nSolution.get(maxi).proposition.size())
					for (int m=0 ; m<nSolution.get(n).chutes.size() ; m++)
						if (nSolution.get(n).chutes.get(m).getSurface() > maxisurf)
							maxi2 = n;
		return maxi2;
	}


}